﻿using LLama;
using LLama.Common;
using LLama.Native;
using LLama.Sampling;
using System.Text.RegularExpressions;
using System.Data.SQLite;
using System.IO;
using System.Windows;

namespace LLM_Doghouse
{
    internal class Llava
    {
        private uint intCtx = 0;
        private float fltTmp = 0;
        private int intGpu = 0;
        private string strMdl = "";
        private string strClp = "";
        private string strSys = "";
        public uint ContextSize 
        {
            get => intCtx; set => intCtx = value;
        }
        public int GpuLayerCount
        {
            get => intGpu; set => intGpu = value;
        }
        public float Temperature
        {
            get => fltTmp; set => fltTmp = value;
        }
        public string ModelPath
        {
            get => strMdl; set => strMdl = value;
        }
        public string ClipPath
        {
            get => strClp; set => strClp = value;
        }
        public string SystemPrompt
        {
            get => strSys; set => strSys = value;
        }
        public Llava()
        {

        }

        public async Task<string> Run(string imgPath)
        {
            try
            {
                string strMsg = "";

                var prompt = $"System: {strSys}\n{{{imgPath}}}\nUser:\n画像に何が写っているか教えてください。\nAssistant:\n";
                var modPara = new ModelParams(strMdl)
                {
                    ContextSize = intCtx,
                    GpuLayerCount = intGpu,
                };
                var llmWeit = LLamaWeights.LoadFromFile(modPara);
                var llmContx = llmWeit.CreateContext(modPara);
                // Llava Init
                using var clipModel = LLavaWeights.LoadFromFile(strClp);
                var itrEx = new InteractiveExecutor(llmContx, clipModel);

                var infPara = new InferenceParams()
                {
                    SamplingPipeline = new DefaultSamplingPipeline()
                    {
                        Temperature = fltTmp,
                        Seed = 1337,
                    },
                    AntiPrompts = ["User:"],
                    MaxTokens = 256
                };

                // Evaluate if we have images
                var imageMatches = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Value);
                var imageCount = imageMatches.Count();
                var hasImages = imageCount > 0;

                if (hasImages)
                {
                    var imagePathsWithCurlyBraces = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Value);
                    var imagePaths = Regex.Matches(prompt, "{([^}]*)}").Select(m => m.Groups[1].Value).ToList();

                    itrEx.Context.NativeHandle.KvCacheRemove(LLamaSeqId.Zero, -1, -1);
                    int index = 0;
                    foreach (var path in imagePathsWithCurlyBraces)
                    {
                        // 最初の画像をタグ<image>に置き換え、残りの画像ではタグを削除します。
                        prompt = prompt.Replace(path, index++ == 0 ? "<image>" : "");
                    }
                    // Initialize Images in executor
                    foreach (var image in imagePaths)
                    {
                        itrEx.Images.Add(File.ReadAllBytes(image));
                    }
                }
                strMsg = "";
                await foreach (var text in itrEx.InferAsync(prompt, infPara))
                {
                    strMsg += text;
                }
                return strMsg;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Llava", MessageBoxButton.OK, MessageBoxImage.Error);
                return "";
            }
        }

    }
}
